;          RGB/YUV conversions using fixed point math

; #########################################################################

      .386                   ; minimum processor needed for 32 bit
      .model flat, stdcall   ; FLAT memory model & STDCALL calling
      option casemap :none   ; set code to case sensitive

; #########################################################################

      include \masm32\include\windows.inc

      include \masm32\include\user32.inc
      include \masm32\include\kernel32.inc

      includelib \masm32\lib\user32.lib
      includelib \masm32\lib\kernel32.lib

      RGB2YUV PROTO :DWORD,:DWORD
      YUV2RGB PROTO :DWORD

YUV STRUCT
   Y  DD  ?
   U  DD  ?
   V  DD  ?
YUV ENDS

.data
; the various factors are initialized with the
; decimals converted to binary fractions

   n114  dd  114*65536/1000  ;0.114 *
   n299  dd  299*65536/1000  ;0.299 *
   n492  dd  492*65536/1000  ;0.492
   n587  dd  587*65536/1000  ;0.587 *
   n877  dd  877*65536/1000  ;0.877
   n1    dd  10000h          ;1.000
   myRGB    dd    00c09050h
   myYUV    YUV <>
   txtbuf   db    20 dup(?)

.code

start:
      mov   ecx,100
   @@:
      push  ecx
      invoke RGB2YUV,myRGB,ADDR myYUV
      invoke YUV2RGB,ADDR myYUV
      mov   myRGB,eax
      pop   ecx
      dec   ecx
      jnz   @B

      mov   ecx,8
      lea   edi,txtbuf
      mov   eax,myRGB
   @@:
      rol   eax,4
      push  eax
      and   al,0fh
      add   al,30h
      .if   al > "9"
            add   al,7
      .endif
      stosb
      pop   eax
      dec   ecx
      jnz   @B
      invoke MessageBox,0,ADDR txtbuf,0,MB_OK
      invoke ExitProcess,0

;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

RGB2YUV proc USES esi clrRGB:DWORD, pYUV:DWORD

LOCAL RED    :DWORD
LOCAL BLUE   :DWORD

   mov  esi,pYUV

;compute Y = 0.299 R + 0.587 G + 0.114 B

   mov  eax,clrRGB
   and  eax,0FFh   ;isolate the RED
   mov  RED,eax
   imul n299
   mov  ecx,eax    ;use ECX as the accumulator

   mov  eax,clrRGB
   shr  eax,8
   and  eax,0FFh   ;isolate the GREEN (will not be reused)
   imul n587
   add  ecx,eax

   mov  eax,clrRGB
   shr  eax,16
   and  eax,0FFh   ;isolate the BLUE
   mov  BLUE,eax
   imul n114
   add  ecx,eax    ;ECX = Y = 0.299 R + 0.587 G + 0.114 B
   mov  [esi],ecx  ;store it

;compute U = 0.492 (B - Y)

   mov  eax,BLUE
   shl  eax,16
   sub  eax,ecx    ;(B - Y)
   imul n492
   shr  eax,16
   shl  edx,16
   add  eax,edx
   mov  [esi+4],eax ;store the U value

;compute V = 0.877 (R - Y)

   mov  eax,RED
   shl  eax,16
   sub  eax,ecx    ;(R - Y)
   imul n877
   shr  eax,16
   shl  edx,16
   add  eax,edx
   mov  [esi+8],eax ;store the V value
   ret
RGB2YUV endp

;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

YUV2RGB proc USES esi pYUV:DWORD

LOCAL RED    :DWORD
LOCAL GREEN  :DWORD
LOCAL BLUE   :DWORD

   mov esi,pYUV

; RED (R = (V/0.877+Y))
   mov  eax,[esi+8]
   imul n1
   idiv n877
   add  eax,[esi]   ;+Y = R
   or   eax,eax     ;test for sign
   jns  @F
   xor  eax,eax     ;negative values can only be due to
                    ;to rounding errors
   jmp storeRED
@@:
   shr  eax,16      ;the most significant bit of the
                    ;fractional part is equivalent to 0.5
                    ;and gets transferred to the CARRY flag
   adc  eax,0       ;increment result if fraction > 0.5

   cmp  eax,255
   jbe  storeRED
   mov  eax,255
storeRED:
   mov  RED,eax

; BLUE (B = (U/0.492+Y))
   mov  eax,[esi+4]
   imul n1
   idiv n492
   add  eax,[esi]   ;+Y = B
   or   eax,eax     ;test for sign
   jns  @F
   xor  eax,eax     ;negative values can only be due to
                    ;to rounding errors
   jmp storeBLUE
@@:
   shr  eax,16      ;the most significant bit of the
                    ;fractional part is equivalent to 0.5
                    ;and gets transferred to the CARRY flag
   adc  eax,0       ;increment result if fraction > 0.5

   cmp  eax,255
   jbe  storeBLUE
   mov  eax,255
storeBLUE:
   mov  BLUE,eax

; GREEN (G = (Y-(0.299R + 0.114B))/0.587)
   mov  ecx,[esi]   ;use ECX as accumulator and
                    ;initialize it with Y
   mov  eax,RED
   mul  n299
   sub  ecx,eax
   mov  eax,BLUE
   mul  n114
   sub  ecx,eax     ;ECX = Y-(0.299R + 0.114B)
   mov  eax,ecx
   cdq              ;extend sign to EDX
   idiv n587
   or   eax,eax     ;test for sign
   jns  @F
   xor  eax,eax     ;negative values can only be due to
                    ;to rounding errors
   jmp storeGREEN
@@:
   shl  dx,1        ;the most significant bit of the
                    ;fractional part is equivalent to 0.5
                    ;and gets transferred to the CARRY flag
   adc  eax,0       ;increment result if fraction > 0.5

   cmp  eax,255
   jbe  storeGREEN
   mov  eax,255
storeGREEN:
   mov  GREEN,eax

   mov  eax,BLUE
   shl  eax,8
   add  eax,GREEN
   shl  eax,8
   add  eax,RED
   ret
YUV2RGB endp

end start
